/***************************************************************************
 *   Copyright (C) 2002-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __MAIN_H__
#define __MAIN_H__

/*****************************************************************************************************************\
 ********************************* INCLUDES **********************************************************************
\*****************************************************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>     /* for catching signals */
#include <time.h>       /* included for logging */
#include <errno.h>      /* error handling */
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/wait.h>

/* our own vuurmuurlib */
#include <vuurmuur.h>

/*****************************************************************************************************************\
 ********************************* DEFINES ***********************************************************************
\*****************************************************************************************************************/

#define VUURMUUR_VERSION_MAJOR		0
#define VUURMUUR_VERSION_MINOR		5
#define VUURMUUR_VERSION_SUB		51

#define MIN_LIBVUURMUUR_VERSION_MAJOR	0
#define MIN_LIBVUURMUUR_VERSION_MINOR	5
#define MIN_LIBVUURMUUR_VERSION_SUB	50

#define MAX_PIPE_COMMAND              512 /* maximum lenght of the pipe command */

#define LOOP_INT                      1

#define YES                           1
#define NO                            0

#define PIDFILE                       "/var/run/vuurmuur.pid"

#define LOCK                          LockSHM(1, sem_id)
#define UNLOCK                        LockSHM(0, sem_id)

/*************************************************************************************************************************\
 ******************************************************* DATATYPES *******************************************************
\*************************************************************************************************************************/

struct RuleCreateData_
{
	char action[sizeof(RuleCache.action)];
	char chain[48]; /* why 48? */

	/* ipaddress */
	char from_ip[16];
	char to_ip[16];

	/* netmasks */
	char from_netmask[16];
	char to_netmask[16];

	/* mac - only from */
	char from_mac[48]; /* --m mac --from-mac mac */

	struct ipdata ipv4_from;
	struct ipdata ipv4_to;

	/* interfaces */
	char from_int[16];
	char to_int[16];

	/* proto */
	char proto[16+6]; // why 16+6? <- the 6 is for ' --syn' for tcp
	char helper[32];

	/* ports */
	char temp_dst_port[32];
	char temp_src_port[32];

	struct portdata *portrange_ptr;
	struct portdata *listenport_ptr;
	struct portdata *remoteport_ptr;

	/* loop stuff - needs a cleanup and needs it BAD */
	int logged;
	int normal;
	int broadcasted;
	int broadlog;

	/* portfw stuff - needs to go-> later we can put it in the function for creating portfw rules! */
	char serverip[MAX_NET_ZONE];
	char remoteip[32]; // max 15 (for ip) + 1 (for :) + 5 (for port) = 21? + 1 = 22
	char temp_port_store[6];
};

/* counters */
int ipt_rulecount;
int ipt_input_count;
int ipt_output_count;
int ipt_forward_count;
int ipt_preroute_count;
int ipt_postroute_count;

/* the services linked-list */
d_list	ServicesList;

/* the zonedata linked-list */
d_list	ZonedataList;

/* the querydata linked-list */
d_list	QuerydataList;

d_list	InterfacesList;

d_list	PluginList;

/* list of ipaddresses to be blocked */
d_list	BlockList;

/* semaphore id */
int sem_id;

/* pointer to the environment */
extern char **environ;

/*************************************************************************************************************************\
 ******************************************************* FUNCTIONS *******************************************************
\*************************************************************************************************************************/

/* rules.c */
int create_all_rules(int debuglvl, d_list *ruleslist, d_list *blocklist, int create_prerules);
int analyze_all_rules(int debuglvl, d_list *ruleslist);

int pre_rules(const int debuglvl, d_list *interfaces_list);
int post_rules(const int debuglvl, int forward_rules);

int update_synlimit_rules(const int debuglvl);
int create_block_rules(const int debuglvl, d_list *blocklist);

int create_rule(int debuglvl, struct RuleCache_ *create);
int remove_rule(const int debuglvl, int chaintype, int first_ipt_rule, int rules);

int create_rule_input(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_output(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_forward(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_masq(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_snat(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_portfw(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);
int create_rule_redirect(const int debuglvl, struct RuleCreateData_ *rule, struct RuleCache_ *create);


/* misc.c */
int pipe_command(const int debuglvl, char *command, char ignore_error);
int vuurmuur_check_lib_version(int debuglvl, int major, int minor, int sub, int min_lib_major, int min_lib_minor, int min_lib_sub);

/* main.c */
// none ;-)

/* reload.c */
int apply_changes(int debuglvl, d_list *services_list, d_list *zones_list, d_list *interfaces_list, d_list *rules_list, d_list *block_list, struct rgx_ *reg);

int reload_services(int debuglvl, d_list *d_list, regex_t *servicename_regex);
int reload_services_check(int debuglvl, struct ServicesData_ *ser_ptr);

int reload_zonedata(int debuglvl, d_list *d_list, struct rgx_ *reg);
int reload_zonedata_check(int debuglvl, d_list *d_list, struct ZoneData_ *zone_ptr, struct rgx_ *reg);

int reload_interfaces(int debuglvl, d_list *d_list);
int reload_interfaces_check(int debuglvl, struct InterfaceData_ *iface_ptr);

#endif
