/***************************************************************************
 *   Copyright (C) 2003-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

/*	pipe_command

	This function takes the 'command' and pipes it to the shell.

	Returncodes:
		 0: ok
		-1: error
 */
int
pipe_command(const int debuglvl, char *command, char ignore_error)
{
	int	retval=0;
	FILE	*p;

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "%s: ** start **", __FUNCTION__);


	/* safety */
	if(!command)
	{
		(void)vrprint.error(-1, "Internal Error", "parameter problem (in: %s:%d).", __FUNCTION__, __LINE__);
		return(-1);
	}

	if(debuglvl >= MEDIUM)
	{
		(void)vrprint.debug("Debug", "pipe_command: command: %s", command);
		(void)vrprint.debug("Debug", "pipe_command: strlen(command) = %d, max = %d", strlen(command), MAX_PIPE_COMMAND);
	}

	if(strlen(command) > MAX_PIPE_COMMAND)
	{
		(void)vrprint.error(-1, "Error", "Command to pipe too long! (%d, while max is: %d).", strlen(command), MAX_PIPE_COMMAND);
		retval=-1;
	}
	else
	{
		/* if in bash output mode we don't pipe, but just print to stdout */
		if(bash_out == 1)
		{
			fprintf(stdout, "%s\n", command);
		}
		else
		{
			if(!(p = popen(command,"r")))
			{
				(void)vrprint.error(-1, "Error", "opening pipe to '%s' failed.", command);
				retval = -1;
			}
			else
			{
				if(debuglvl >= MEDIUM)
					(void)vrprint.debug("Debug", "pipe_command: Pipe opened succesfully.");

				if(pclose(p))
				{
					if(!ignore_error)
					{
						(void)vrprint.error(-1, "Error", "closing pipe: '%s' failed.", command);
						retval = -1;
					}
				}
				else
				{
					if(debuglvl >= MEDIUM)
						(void)vrprint.debug("Debug", "pipe_command: Pipe closed!");
				}
			}
		}
	}

	if(debuglvl >= MEDIUM)
		(void)vrprint.debug("Debug", "pipe_command: ** end **, return=%d", retval);

	return(retval);
}

/*	vuurmuur_check_lib_version

	This function checks the version of libvuurmuur.

		 0: ok
		-1: error
*/
int
vuurmuur_check_lib_version(int debuglvl, int major, int minor, int sub, int min_lib_major, int min_lib_minor, int min_lib_sub)
{
	int	retval=0;
	int	lib_major,
		lib_minor,
		lib_sub;
	int	min_vuur_major,
		min_vuur_minor,
		min_vuur_sub;

	tell_libvuurmuur_version(&lib_major, &lib_minor, &lib_sub, &min_vuur_major, &min_vuur_minor, &min_vuur_sub);

	if(lib_major >= min_lib_major && lib_minor >= min_lib_minor && lib_sub >= min_lib_sub)
	{

	}
	else
	{
		fprintf(stdout, "Libvuurmuur version: %d.%d.%d, ", lib_major, lib_minor, lib_sub);
		fprintf(stdout, "minimal required version: %d.%d.%d: ", min_lib_major, min_lib_minor, min_lib_sub);
		fprintf(stdout, "error, update your libvuurmuur.\n");

		retval=-1;
	}


	if(major >= min_vuur_major && minor >= min_vuur_minor && sub >= min_vuur_sub)
	{

	}
	else
	{
		fprintf(stdout, "Vuurmuur version: %d.%d.%d, ", major, minor, sub);
		fprintf(stdout, "minimal required version (by libvuurmuur): %d.%d.%d, ", min_vuur_major, min_vuur_minor, min_vuur_sub);
		fprintf(stdout, "error, you need to update Vuurmuur.\n");

		retval=-1;
	}

	return(retval);
}
