/***************************************************************************
 *   Copyright (C) 2002-2004 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"

/* we put this here, because we only use it here in main. */
static int sigint_count = 0;
static int sighup_count = 0;
static void catch_sigint(int signo);
static void catch_sighup(int signo);

/*@null@*/
struct SHM_TABLE *shm_table = 0;

int
main(int argc, char *argv[])
{
	int			retval=0,
				optch,
				result=0,
				loop=0,
				nodaemon=0,
				debuglvl = 0;
	uid_t			user;
	gid_t			group;
	static char		optstring[] = "hd:bVlvnc:L:";
	int			reload=0;
	struct stat		stat_buf;
	int			shm_id;
	char			*shmp;
	union semun		semarg;
	ushort			seminit[] = { 1,0 };

	int			wait_time = 0;	// time in seconds we have waited for an VR_RR_RESULT_ACK when using SHM-IPC
	struct rgx_		reg;		// regexes

	// registering signals we use
	if(signal(SIGINT, &catch_sigint) == SIG_ERR)
	{
		fprintf(stdout, "Error: couldn't attach the signal SIGINT to the signal handler.\n");
		retval=-1;
	}
	if(signal(SIGHUP, &catch_sighup) == SIG_ERR)
	{
		fprintf(stdout, "Error: couldn't attach the signal SIGHUP to the signal handler.\n");
		retval=-1;
	}

	libvuurmuur_verbose=0, bash_out=0;
	conf.loglevel_cmdline = 0;
	strlcpy(conf.configfile_location, VUURMUUR_CONFIGFILE, sizeof(conf.configfile_location));

	// Process commandline options
	while((optch = getopt(argc, argv, optstring)) != -1 )
	{
		switch(optch)
		{
			case 'c' :
				// config file
				if(libvuurmuur_verbose == 1)
					fprintf(stdout, "Using this configfile: %s\n", optarg);

				if(strlcpy(conf.configfile_location, optarg, sizeof(conf.configfile_location)) > sizeof(conf.configfile_location))
				{
					fprintf(stdout, "Error: configfile (-c): argument too long (max: %d).\n", sizeof(conf.configfile_location)-1);
					exit(EXIT_FAILURE);
				}
				break;

			case 'd' :
				// debugging
				fprintf(stdout, "vuurmuur: debugging enabled.\n");

				// convert the debug string and check the result
				debuglvl = atoi(optarg);
				if(debuglvl < 0 || debuglvl > HIGH)
				{
					fprintf(stdout, "Error: illegal debug level: %d (max: %d).\n", debuglvl, HIGH);
					exit(EXIT_FAILURE);
				}

				fprintf(stdout, "vuurmuur: debug level: %d\n", debuglvl);
				libvuurmuur_debug = debuglvl;
				break;

			case 'L' :

				conf.loglevel_cmdline = 1;

				/* loglevel */
				if(strlcpy(conf.loglevel, optarg, sizeof(conf.loglevel)) > sizeof(conf.loglevel))
				{
					fprintf(stdout, "Error: loglevel (-L): argument too long (max: %d).\n", sizeof(conf.loglevel)-1);
					exit(EXIT_FAILURE);
				}

				break;

			case 'b' :
				// bash output
				bash_out=1;
				fprintf(stdout, "#!/bin/sh\n");
				fprintf(stdout, "# Firewall generated by Vuurmuur %d.%d.%d, Copyright (C) 2002-2004 by Victor Julien\n", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
				break;

			case 'h' :
				// help
				fprintf(stdout, "Usage: vuurmuur [OPTION]\n");
				fprintf(stdout, "\n");
				fprintf(stdout, "Options:\n");
				fprintf(stdout, "-b gives a bashscript output\n");
				fprintf(stdout, "-d [1 - 3] enables debugging, 1 low, 3 high\n");
				fprintf(stdout, "-h gives this help\n");
				fprintf(stdout, "-V gives the version\n");
				fprintf(stdout, "-l vuurmuur starts and goes into daemon-mode.\n");
				fprintf(stdout, "-v verbose mode.\n");
				fprintf(stdout, "-n for use with -l, it goes into the loop without daemonizing.\n");
				fprintf(stdout, "\n");
				exit(0);

			case 'V' :
				// print version
				fprintf(stdout, "Vuurmuur %d.%d.%d\n", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
				fprintf(stdout, "Copyright (C) 2002-2004 by Victor Julien\n");
				exit(0);

			case 'l' :
				// looping
				loop=1;
				break;

			case 'v' :
				// verbose
				fprintf(stdout, "verbose output\n");
				libvuurmuur_verbose=1;
				break;

			case 'n' :
				// dont daemonize
				fprintf(stdout, "no daemon\n");
				nodaemon=1;
				break;

			default:
				//fprintf(stdout, "Error: unknown option '-%c'. See -h for valid options.\n", optch);
				break;
		}
	}

//TODO: do an options sanity check (eg. bash and loop dont play together)


	// check if were already running, but not if we want bash output
	if(bash_out == 0)
	{
		if(check_pidfile(PIDFILE) == -1)
			exit(-1);
	}

	/* check if the binary is a symlink */
	if(lstat(argv[0], &stat_buf) == -1)
	{
		fprintf(stdout, "Error: lstat failed on the executable.\n");
		return(-1);
	}
	else
	{
		if(S_ISLNK(stat_buf.st_mode) == 1)
		{
			fprintf(stdout, "For security reasons Vuurmuur will not run when called from a symbolic link.\n");
			return(-1);
		}
	}

	/* init the print functions: all to stdout */
	vrprint.error = libvuurmuur_stdoutprint_error;
	vrprint.warning = libvuurmuur_stdoutprint_warning;
	vrprint.info = libvuurmuur_stdoutprint_info;
	vrprint.debug = libvuurmuur_stdoutprint_debug;

	/* get userid and groupid */
	user = getuid();
	group = getgid();

	if(debuglvl >= 1)
	{
		fprintf(stdout, "main: user=%d, group=%d\n", (int)user, (int)group);
	}

	/*	exit if the user is not root. BUG/TODO - we could enable this program also for non-root users,
		but for example only with bashout on and loop off...
	*/
	if(user > 0 || group > 0)
	{
		fprintf(stdout, "Error: you are not root! Exitting.\n");
		exit(EXIT_FAILURE);
	}

	/* check for the version of the library */
	result = vuurmuur_check_lib_version(debuglvl, VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB, MIN_LIBVUURMUUR_VERSION_MAJOR, MIN_LIBVUURMUUR_VERSION_MINOR, MIN_LIBVUURMUUR_VERSION_SUB);
	if(result < 0)
		exit(EXIT_FAILURE);

	/* init the config */
	if(retval == 0)
	{
		if(debuglvl >= MEDIUM)
			(void)vrprint.debug("Debug", "main: initializing config... calling init_config()");

		result = init_config(debuglvl, conf.configfile_location, &conf);
		if (result == 0)
		{
			if(debuglvl >= MEDIUM)
				(void)vrprint.debug("Debug", "main: initializing config complete and succesful.");
		}
		else
		{
			fprintf(stdout, "Initializing config failed.\n");
			exit(EXIT_FAILURE);
		}
	}

	/* now we know the logfile locations, so init the log functions */
	vrprint.error = libvuurmuur_logprint_error;
	vrprint.warning = libvuurmuur_logprint_warning;
	vrprint.info = libvuurmuur_logprint_info;
	vrprint.debug = libvuurmuur_logprint_debug;

	/* setup regexes */
	if(setup_rgx(1, &reg) < 0)
	{
		(void)vrprint.error(-1, "Internal Error", "setting up regular expressions failed.");
		exit(EXIT_FAILURE);
	}

	/* load the backends */
	result = load_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		fprintf(stdout, "Error: loading backends failed\n");
		exit(EXIT_FAILURE);
	}

	/* print some nice info about me being the coolest of 'm all ;-) */
	(void)vrprint.info("Info", "This is Vuurmuur %d.%d.%d", VUURMUUR_VERSION_MAJOR, VUURMUUR_VERSION_MINOR, VUURMUUR_VERSION_SUB);
	(void)vrprint.info("Info", "Copyright (C) 2002-2004 by Victor Julien");

	/* set chain couters to zero */
	ipt_rulecount=0;
	ipt_input_count=0;
	ipt_output_count=0;
	ipt_forward_count=0;
	ipt_preroute_count=0;
	ipt_postroute_count=0;

	/* load the services into memory */
	(void)vrprint.info("Info", "Loading services...");
	result = init_services(debuglvl, &ServicesList, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading services succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all services where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "init_services failed.");
		exit(EXIT_FAILURE);
	}
	//services_print_list(&ServicesList);

	/* load the interfaces into memory */
	result = init_interfaces(debuglvl, &InterfacesList);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading interfaces succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all interfaces where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "init_interfaces failed.");
		exit(EXIT_FAILURE);
	}

	/* load the zonedata into memory */
	(void)vrprint.info("Info", "Loading zones...");
	result = init_zonedata(debuglvl, &ZonedataList, &InterfacesList, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading zonedata succesfull.");
	}
	else if(result == 1)
	{
		(void)vrprint.warning("Warning", "Not all zones where loaded!");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "init_zonedata failed.");
		exit(EXIT_FAILURE);
	}
	//zonedata_print_list(&ZonedataList);


	/* load the blockfile if any */
	if(conf.use_blocklist)
	{
		/* call it with load_ips == 1 */
		if(blocklist_read_file(debuglvl, &ZonedataList, &BlockList, 1) < 0)
			conf.use_blocklist = 0;
	}


	/* load the rulesfile into memory */
	(void)vrprint.info("Info", "Loading rulesfile...");

	result = rules_init_list(debuglvl, &QuerydataList, &reg);
	if(result == 0)
	{
		(void)vrprint.info("Info", "Loading rulesfile succesfull.");
	}
	else
	{
		(void)vrprint.error(-1, "Error", "init_rulesfile failed.");
		exit(EXIT_FAILURE);
	}

	/* analyzing the rules */
	if(analyze_all_rules(debuglvl, &QuerydataList) != 0)
	{
		(void)vrprint.error(-1, "Error", "analizing rules failed.");
		exit(EXIT_FAILURE);
	}

	if(debuglvl >= LOW)
		query_print_list(&QuerydataList);

	/* now create the rules */
	if(create_all_rules(debuglvl, &QuerydataList, &BlockList, 1) != 0)
	{
		(void)vrprint.error(-1, "Error", "creating rules failed.");
		exit(EXIT_FAILURE);
	}

// ** analyzing and creating rules done ** //

	/* enter the infinite loop... */
	if(loop == 1 && retval == 0)
	{
		// if we going to enter the loop, make sure we dont are in bash-mode
		if(bash_out == 0)
		{
			// leave console
			if(nodaemon == 0)
			{
				if(daemon(1,1) < 0)
				{
					(void)vrprint.error(-1, "Error", "failed to daemonize: %s.", strerror(errno));
					exit(EXIT_FAILURE);
				}
				else
					(void)vrprint.info("Info", "Entered daemon mode: obtained a new PID (%ld).", getpid());
			}

			shm_id = shmget(IPC_PRIVATE, sizeof(*shm_table), 0600);
			if(shm_id < 0)
			{
				(void)vrprint.error(-1, "Error", "Unable to create shared memory: %s.", strerror(errno));
				exit(EXIT_FAILURE);
			}
			/* for some reason on my machine the shm_id is zero when vuurmuur is started at boot
			   if we sleep for some time and retry it works */
			else if(shm_id == 0)
			{
				/* sleep 3 seconds before trying again */
				(void)sleep(3);

				shm_id = shmget(IPC_PRIVATE, sizeof(*shm_table), 0600);
				if(shm_id < 0)
				{
					(void)vrprint.error(-1, "Error", "Unable to create shared memory: %s (retry).", strerror(errno));
					exit(EXIT_FAILURE);
				}
				else if(shm_id == 0)
				{
					(void)vrprint.info("Info", "Still no valid shm_id. Giving up.");
				}
				else
				{
					(void)vrprint.info("Info", "Creating shared memory succesfull: shm_id: %d, 0x08%lX, 0x08%lX (retry).", shm_id, (long)shm_table, (long)shm_table+sizeof(*shm_table));
				}
			}
			else
			{
				(void)vrprint.info("Info", "Creating shared memory succesfull: shm_id: %d.", shm_id);
			}

			/* now attach to the shared mem */
			if(shm_id > 0)
			{
				shmp = shmat(shm_id, 0, 0);
				if(shmp == (char *)(-1))
				{
					shm_table = NULL;
					(void)vrprint.error(-1, "Error", "Unable to attach to shared memory: %s.", strerror(errno));
				}
				else
				{
					shm_table = (struct SHM_TABLE *)shmp;
					(void)vrprint.info("Info", "Attaching to shared memory successfull.");
				}

				/* if all went well we create a semaphore */
				if(shm_table)
				{
					sem_id = semget(IPC_PRIVATE, 2, 0600);
					if(sem_id == -1)
					{
						(void)vrprint.error(-1, "Error", "Unable to create semaphore: %s.", strerror(errno));
						exit(EXIT_FAILURE);
					}
					else
					{
						(void)vrprint.info("Info", "Creating a semaphore success: %d", sem_id);
					}

					semarg.array = seminit;
					if(semctl(sem_id, 0, SETALL, semarg) == -1)
					{
						(void)vrprint.error(-1, "Error", "Unable to initialize semaphore: %s.", strerror(errno));
						exit(EXIT_FAILURE);
					}
					else
					{
						(void)vrprint.info("Info", "Initializeing the semaphore successfull.");
					}

					/* now initialize the shared mem */
					if(LOCK)
					{
						shm_table->sem_id = sem_id;
						shm_table->backend_changed = 0;
						shm_table->reload_result = VR_RR_READY;

						UNLOCK;
					}
				}
			}

			/* create a pidfile */
			result = create_pidfile(PIDFILE, shm_id);
			if(result < 0)
			{
				(void)vrprint.error(-1, "Error", "Unable to create pidfile.");
/* TODO: is this really that serious? */
				exit(EXIT_FAILURE);
			}

			(void)vrprint.info("Info", "Entering the loop... (interval %d seconds)", LOOP_INT);

			while(retval == 0 && sigint_count == 0)
			{
				if(LOCK)
				{
					if(shm_table->configtool.connected == 1)
					{
						(void)vrprint.info("Info", "Configtool connected: %s.", shm_table->configtool.name);
						shm_table->configtool.connected=2;
					}
					else if(shm_table->configtool.connected == 3)
					{
						(void)vrprint.info("Info", "Configtool disconnected: %s.", shm_table->configtool.name);
						shm_table->configtool.connected=0;
					}

					if(shm_table->backend_changed == 1)
					{
						(void)vrprint.info("Info", "IPC-SHM: backend changed.");
						reload = 1;
						shm_table->backend_changed = 0;
					}

					UNLOCK;
				}

				/* well, we either recieved a SIGHUP or we want to reload trough an IPC command. */
				if(sighup_count == 1 || reload == 1)
				{
					/* apply changes */
					result = apply_changes(debuglvl, &ServicesList, &ZonedataList, &InterfacesList, &QuerydataList, &BlockList, &reg);
					if(result < 0)
					{
						(void)vrprint.error(-1, "Error", "Applying changes failed.");
/* TODO: shouldn't we use retval here? */
						exit(EXIT_FAILURE);
					}
					else
					{
						sighup_count = 0;
					}

					/* if we are reloading because of an IPC command, we need to communicate with the caller */
					if(reload == 1)
					{
						if(LOCK)
						{
							/* tell the caller about the reload result */
							if(result < 0)
							{
								shm_table->reload_result = VR_RR_ERROR;
							}
							else if(result == 0)
							{
								shm_table->reload_result = VR_RR_SUCCES;
							}
							else
							{
								shm_table->reload_result = VR_RR_NOCHANGES;
							}
							UNLOCK;
						}
						reload = 0;
					
						(void)vrprint.info("Info", "Waiting for an VR_RR_RESULT_ACK");

						result = 0;
						wait_time = 0;

						/* now wait max 30 seconds for an ACK from the caller */
						while(result == 0 && wait_time < 30)
						{
							if(LOCK)
							{
								/* ah, we got one */
								if(shm_table->reload_result == VR_RR_RESULT_ACK)
								{
									shm_table->reload_result = VR_RR_READY;
									result = 1;

									(void)vrprint.info("Info", "We got an VR_RR_RESULT_ACK!");
								}
								UNLOCK;
							}

							wait_time++;
							sleep(1);
						}

						/* damn, we didn't get one */
						if(result == 0)
						{
							(void)vrprint.info("Info", "We've waited for %d seconds for an VR_RR_RESULT_ACK, but got none. Setting to VR_RR_READY", wait_time);
							if(LOCK)
							{
								shm_table->reload_result = VR_RR_READY;
								UNLOCK;
							}
							else
							{
								(void)vrprint.info("Info", "Hmmmm, failed to set to ready. Did the client crash?");
							}
						}
						result = 0;
					}
				}

				sleep(LOOP_INT);
			}

			(void)vrprint.info("Info", "Destroying shared memory...");
			if(shmctl(shm_id, IPC_RMID, NULL) < 0)
			{
				fprintf(stdout, "Error shmctl: %s.\n", strerror(errno));
				exit(-3);
			}
			else
			{
				(void)vrprint.info("Info", "Shared memory destroyed.");
			}

			/* destroy semaphore */
			if(semctl(sem_id, 0, IPC_RMID, semarg) == -1)
			{
				(void)vrprint.error(-1, "Error", "Failed to remove semaphore.");
				retval=-1;
			}

			/* remove the pidfile */
			if(remove_pidfile(PIDFILE) < 0)
			{
				(void)vrprint.error(-1, "Error", "Unable to remove pidfile: %s.", strerror(errno));
				retval=-1;
			}

			(void)vrprint.info("Info", "Loop shutting down...");
		}
		else
		{
			fprintf(stdout, "# loop mode not supported when using -b (bash output).\n");
		}
	}

	/* unload the backends */
	result = unload_backends(debuglvl, &PluginList);
	if(result < 0)
	{
		fprintf(stdout, "Error: unloading backends failed.\n");
		exit(EXIT_FAILURE);
	}


	/*
		Destroy the linked-lists
	*/

	/* destroy the ServicesList */
	destroy_serviceslist(debuglvl, &ServicesList);

	/* destroy the ZonedataList */
	destroy_zonedatalist(debuglvl, &ZonedataList);

	/* destroy the InterfacesList */
	destroy_interfaceslist(debuglvl, &InterfacesList);

	/* destroy the QuerydataList */
	if(rules_cleanup_list(debuglvl, &QuerydataList) < 0)
		retval = -1;
		
	d_list_cleanup(debuglvl, &BlockList);

	/* cleanup regexes */
	(void)setup_rgx(0, &reg);

	if(debuglvl >= HIGH)
		(void)vrprint.debug("Debug", "main: ** end **, return = %d", retval);

	return(retval);
}

static void catch_sigint(int signo)
{
	sigint_count=1;
}

static void catch_sighup(int signo)
{
	sighup_count=1;
}
